from .accept_invite import *
from .activity_log import *
from .add import *
from .change_log import *
from .dashboard import *
from .delete import *
from .list import *
from .remove_lock import *
from .request_new_invite import *
from .resend_invite import *
from .reset_password import *
from .revoke_session import *
from .security import *
from .set_new_password import *
from .sign_in import *
from .sign_out import *
from .update import *
from .update_my_profile import *
from .view import *

from . import mfa



"""
Placeholder dashboard view.
"""

from manhattan.chains import Chain, ChainMgr
from manhattan.manage.views import factories, utils
from manhattan.nav import Nav, NavItem

__all__ = ['dashboard_chains']


# Chains
dashboard_chains = ChainMgr()

# GET
dashboard_chains['get'] = Chain([
    'authenticate',
    'decorate',
    'render_template'
])

# Define the links
dashboard_chains.set_link(factories.authenticate())
dashboard_chains.set_link(factories.render_template('dashboard.html'))

@dashboard_chains.link
def decorate(state):
    state.decor = utils.base_decor(state.manage_config, 'dashboard')
    state.decor['title'] = 'Dashboard'
    state.decor['breadcrumbs'].add(NavItem('Dashboard', ''))



"""
View the user's recovery codes.
"""

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.manage.views import factories, utils
from manhattan.nav import NavItem

__all__ = ['recovery_codes_chains']


# Chains
recovery_codes_chains = ChainMgr()

# POST
recovery_codes_chains['get'] = Chain([
    'config',
    'authenticate',
    'get_recovery_codes',
    'decorate',
    'render_template'
])

# Define the links
recovery_codes_chains.set_link(factories.config(open_user_nav=True))
recovery_codes_chains.set_link(factories.authenticate())
recovery_codes_chains.set_link(
    factories.render_template('mfa/recovery_codes.html')
)

@recovery_codes_chains.link
def decorate(state):
    state.decor = utils.base_decor(state.manage_config, 'mfa_recovery_codes')
    state.decor['title'] = 'Recovery codes'

    # Breadcrumb
    state.decor['breadcrumbs'].add(
        NavItem('Security', state.manage_config.get_endpoint('security'))
    )
    state.decor['breadcrumbs'].add(NavItem('Recovery codes', ''))

    # Actions

    state.decor['actions'].add(
        NavItem(
            'Copy',
            fixed_url='#copy-recovery-codes',
            data={'data-mh-copy-to-clipboard': '[data-mh-recovery-codes]'}
        )
    )

    state.decor['actions'].add(
        NavItem(
            'Download',
            state.manage_config.get_endpoint('mfa_download_recovery_codes')
        )
    )

    state.decor['actions'].add(
        NavItem(
            'Regenerate',
            fixed_url='#regenerate-recovery-codes',
            data={
                'data-mh-submit-by-proxy': '#regenerate-recovery-codes-form'
            }
        )
    )

@recovery_codes_chains.link
def get_recovery_codes(state):
    # Add the recovery codes to the state
    user_cls = state.manage_config.frame_cls
    user = getattr(flask.g, user_cls.get_g_key())

    state.mfa_recovery_codes = user.mfa_recovery_codes



"""
Allow users to request a new invite.
"""

from datetime import datetime
import os

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.formatters.flask import path_to_url
from manhattan.forms import BaseForm, fields, validators
from manhattan.manage.views import factories, utils
from manhattan.secure import random_token
from mongoframes import Q

__all__ = ['request_new_invite_chains']


# Forms

class RequestNewInviteForm(BaseForm):

    email = fields.StringField(
        'Email',
        [
            validators.Required(),
            validators.Email(),
            validators.Length(max=254)
        ]
    )


# Chains
request_new_invite_chains = ChainMgr()

# GET
request_new_invite_chains['get'] = Chain([
    'config',
    'init_form',
    'decorate',
    'render_template'
    ])

# POST
request_new_invite_chains['post'] = Chain([
    'config',
    'init_form',
    'validate',
    [
        [
            'get_user',
            'limit_rate',
            'send_new_invite',
            'redirect'
        ], [
            'decorate',
            'render_template'
        ]
    ]
])

# Define the links
request_new_invite_chains.set_link(
    factories.config(form_cls=RequestNewInviteForm)
)
request_new_invite_chains.set_link(factories.validate())
request_new_invite_chains.set_link(
    factories.render_template('request_new_invite.html')
)
request_new_invite_chains.set_link(factories.redirect('request_new_invite'))

@request_new_invite_chains.link
def decorate(state):
    state.decor = utils.base_decor(state.manage_config, 'request_new_invite')
    state.decor['title'] = 'Request new invite'

@request_new_invite_chains.link
def init_form(state):
    # Initialize the form
    state.form = state.form_cls(flask.request.form)

    # Store the user class against the form
    state.form._user_cls = state.manage_config.frame_cls

@request_new_invite_chains.link
def get_user(state):
    """Get the user to send a new invite to"""
    user_cls = state.manage_config.frame_cls
    user = user_cls.one(Q.email_lower == state.form.data['email'].lower())
    state[state.manage_config.var_name] = user

    if not user or user.invite_accepted:
        # If we couldn't find a user based on the email address the user gave
        # then we don't indicate this to the user as this can be exploited to
        # determine registered email addresses.
        flask.flash('A new invite has been emailed to you.')

        # Redirect the user back to the request new invite page
        return flask.redirect(
            flask.url_for(
                state.manage_config.get_endpoint('request_new_invite')
            )
        )

@request_new_invite_chains.link
def limit_rate(state):
    """Limit the number of reset requests a user can make to once a minute"""
    user = state[state.manage_config.var_name]

    if user.invited is None:
        # User doesn't have an invite pending no need to rate limit
        return

    if (datetime.utcnow() - user.invited).total_seconds() < 60:

        # Let the user know an email has been sent to them but don't send one
        # as we did only 60 seconds ago.
        #
        # NOTE: We used to flag to users that they should wait but this can be
        # exploited to determine registered email addresses and so we no
        # longer provide this feedback.
        #
        flask.flash('A new invite has been emailed to you.')

        # Redirect the user back to the request new invite page
        return flask.redirect(
            flask.url_for(
                state.manage_config.get_endpoint('request_new_invite')
            )
        )

@request_new_invite_chains.link
def send_new_invite(state):
    """Send the user a new invite"""

    env = flask.current_app.jinja_env
    template_path = os.path.join(
        state.manage_config.template_path,
        'emails/invite.html'
    )
    user = state[state.manage_config.var_name]

    # Assign the user an invite token
    user.invited = datetime.utcnow()
    user.invite_token = random_token()
    user.update('invited', 'invite_token')

    # Send the user an invite by email
    flask.current_app.manage.send_email(
        [user.email],
        'Your invite',
        template_path,
        global_vars={
            'user': user.to_json_type(),
            'invite_link': path_to_url(
                flask.url_for(
                    state.manage_config.get_endpoint('accept_invite'),
                    invite_token=user.invite_token
                )
            )
        },
        template_map={
            'invite': env.loader.get_source(
                env,
                'manhattan/users/emails/invite.html'
            )[0],
            template_path: env.loader.get_source(env, template_path)[0]
        }
    )

    # Let the user know to expect an email containing their reset password
    # instructions.
    flask.flash('A new invite has been emailed to you.')



"""
Enable multi-factor authentication (MFA) for the user.
"""

from datetime import datetime
import io

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.forms import BaseForm, fields, validators
from manhattan.manage.views import factories, utils
from manhattan.nav import NavItem
from manhattan.secure import random_code
import pyotp
import qrcode
import qrcode.image.svg

__all__ = ['enable_chains']


# Forms

class EnableForm(BaseForm):

    otp_secret = fields.HiddenField()

    totp = fields.StringField(
        'Your six-digit code',
        [validators.Required()],
        render_kw={'autofocus': True}
    )

    def validate_totp(form, field):

        if not field.data:
            return

        topt = field.data.replace(' ', '')
        if not pyotp.TOTP(form.otp_secret.data).verify(topt):
            raise validators.ValidationError('Not a valid code')


# Chains
enable_chains = ChainMgr()

# POST
enable_chains['get'] = Chain([
    'config',
    'authenticate',
    'init_form',
    'generate_qr_code',
    'decorate',
    'render_template'
])

# POST
enable_chains['post'] = Chain([
    'config',
    'authenticate',
    'init_form',
    'generate_qr_code',
    'validate',
    [
        [
            'enable_mfa',
            'redirect'
        ], [
            'decorate',
            'render_template'
        ]
    ]
])

# Define the links
enable_chains.set_link(
    factories.config(
        form_cls=EnableForm,
        open_user_nav=True
    )
)
enable_chains.set_link(factories.authenticate(mfa_required=False))
enable_chains.set_link(factories.init_form())
enable_chains.set_link(factories.validate())
enable_chains.set_link(factories.render_template('mfa/enable.html'))
enable_chains.set_link(factories.redirect('mfa_recovery_codes'))

@enable_chains.link
def decorate(state):
    state.decor = utils.base_decor(state.manage_config, 'mfa_enable')
    state.decor['title'] = 'Enable 2-factor authentication'

    # Breadcumbs
    state.decor['breadcrumbs'].add(
        NavItem('Security', state.manage_config.get_endpoint('security'))
    )
    state.decor['breadcrumbs'].add(
        NavItem('Enable 2-factor authentication', '')
    )

@enable_chains.link
def generate_qr_code(state):
    """
    Generate a secret and QR code to allow a user to verify they have enabled
    QR on their device.
    """
    user_cls = state.manage_config.frame_cls
    user = getattr(flask.g, user_cls.get_g_key())
    prefix = user_cls.get_settings_prefix()

    # Make sure we have a secret for generating time-based one time passwords
    state.otp_secret = state.form.data.get('otp_secret') \
            or pyotp.random_base32()

    state.form.otp_secret.data = state.otp_secret

    # Generate the provisioning URL
    mfa_client = pyotp.totp.TOTP(state.otp_secret)
    mfa_provisision_uri = mfa_client.provisioning_uri(
        user.email,
        issuer_name=flask.current_app.config[f'{prefix}_MFA_ISSUER']
    )

    # Generate the QR code as an SVG
    f = io.BytesIO()
    qrcode.make(
        mfa_provisision_uri,
        image_factory=qrcode.image.svg.SvgPathImage
    ).save(f)
    f.seek(0)

    state.qr_image = f.read().decode('utf8')

@enable_chains.link
def enable_mfa(state):
    """Enable multi-factor authentication for the user"""
    user_cls = state.manage_config.frame_cls
    user = getattr(flask.g, user_cls.get_g_key())
    prefix = user_cls.get_settings_prefix()

    # Generate recovery codes
    recovery_codes = []
    recovery_code_count = flask\
        .current_app\
        .config[f'{prefix}_MFA_RECOVERY_CODES']

    for i in range(0, recovery_code_count):
        recovery_codes.append(
            random_code(
                flask.current_app.config[f'{prefix}_MFA_RECOVERY_CODE_LENGTH']
            )
        )

    # Enable MFA fo
    user.logged_update(
        user,
        {
            'mfa_enabled': True,
            'mfa_otp_secret': state.otp_secret,
            'mfa_recovery_codes': recovery_codes
        },
        'mfa_enabled',
        'mfa_otp_secret',
        'mfa_recovery_codes',
        'modified'
    )

    # Update the users sessions MFA authorized date/time
    flask.current_session = user.current_session
    flask.current_session.mfa_authorized = datetime.utcnow()
    flask.current_session.update('mfa_authorized')



"""
Remove a lock from a user.
"""

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.manage.views import factories, generic

__all__ = ['remove_lock_chains']


# Chains
remove_lock_chains = ChainMgr()

# POST
remove_lock_chains['post'] = Chain([
    'config',
    'authenticate',
    'get_document',
    'remove_lockout',
    'redirect'
])

# Define the links
remove_lock_chains.set_link(factories.config(projection={'email': True}))
remove_lock_chains.set_link(factories.authenticate())
remove_lock_chains.set_link(factories.get_document())
remove_lock_chains.set_link(factories.redirect('view', include_id=True))

@remove_lock_chains.link
def remove_lockout(state):
    user_cls = state.manage_config.frame_cls
    user = state[state.manage_config.var_name]
    prefix = user_cls.get_settings_prefix()

    cache = flask.current_app.config[f'{prefix}_FAILED_SIGN_IN_CACHE']
    cache.delete(user_cls.get_sign_in_attempt_key(user.email))

    flask.flash('Lock removed')



"""
Add a user.
"""

from datetime import datetime
import os

import flask
from manhattan.formatters.flask import path_to_url
from manhattan.forms import BaseForm, fields, validators
from manhattan.manage.views import factories, generic
from manhattan.secure import random_token
from mongoframes import Q

__all__ = ['add_chains']


# Forms

class AddForm(BaseForm):

    first_name = fields.StringField(
        'First name',
        validators=[
            validators.Required(),
            validators.Length(max=20)
            ]
        )

    last_name = fields.StringField(
        'Last name',
        validators=[
            validators.Required(),
            validators.Length(max=20)
            ]
        )

    email = fields.StringField(
        'Email',
        validators=[
            validators.Required(),
            validators.Email(),
            validators.Length(max=80)
            ]
        )

    def validate_email(form, field):
        """Validate email is unique"""
        user_cls = form._user_cls
        user = user_cls.one(
            Q.email_lower == field.data.lower(),
            projection={'_id': True}
        )

        if user:
            raise validators.ValidationError(
                'This email address is already registered.'
            )


# Chains
add_chains = generic.add.copy()
add_chains['post'].insert_link('add_document', 'invite', after=True)

# Factory overrides
add_chains.set_link(factories.config(form_cls=AddForm))

# Custom overrides

@add_chains.link
def init_form(state):
    generic.add['get'].super(state)

    # Store the user class against the form
    state.form._user_cls = state.manage_config.frame_cls

@add_chains.link
def invite(state):
    """
    Create an invite link for the user and send them an email containing the
    link.
    """

    env = flask.current_app.jinja_env
    template_path = os.path.join(
        state.manage_config.template_path,
        'emails/invite.html'
    )
    user = state[state.manage_config.var_name]

    # Assign the user an invite token
    user.invited = datetime.utcnow()
    user.invite_token = random_token()
    user.update('invited', 'invite_token')

    # Send the user an invite by email
    flask.current_app.manage.send_email(
        [user.email],
        'Your invite',
        template_path,
        global_vars={
            'user': user.to_json_type(),
            'invite_link': path_to_url(
                flask.url_for(
                    state.manage_config.get_endpoint('accept_invite'),
                    invite_token=user.invite_token
                )
            )
        },
        template_map={
            'invite': env.loader.get_source(
                env,
                'manhattan/users/emails/invite.html'
            )[0],
            template_path: env.loader.get_source(env, template_path)[0]
        }
    )



"""
View a user.
"""

from manhattan.manage.views import factories, generic
from manhattan.nav import NavItem
from mongoframes import Q, SortBy

__all__ = ['view_chains']


# Chains
view_chains = generic.view.copy()
view_chains['get'].insert_link('get_document', 'get_sessions', after=True)

# Custom overrides

@view_chains.link
def decorate(state):
    generic.view.chains['get'].get_link('decorate')(state)

    # Actions
    user_cls = state.manage_config.frame_cls
    user = state[state.manage_config.var_name]

    if not user.invite_accepted:

        # Reinvite
        state.decor['actions'].add(
            NavItem(
                'Re-invite',
                fixed_url='#resend-invite',
                data={'data-mh-submit-by-proxy': '#resend-invite-form'}
            ),
            index=0
        )

    if user_cls.is_locked_out(user.email):

        # Remove lockout
        state.decor['actions'].add(
            NavItem(
                'Remove lock',
                fixed_url='#remove-lock',
                data={'data-mh-submit-by-proxy': '#remove-lock-form'}
            ),
            index=0
        )

@view_chains.link
def get_sessions(state):
    """Get a list of sessions for the user"""

    user_cls = state.manage_config.frame_cls
    user = state[state.manage_config.var_name]

    state.user_sessions = user_cls.get_session_cls().many(
        Q.user == user,
        sort=SortBy(Q.last_accessed.desc)
    )




from datetime import datetime
import secrets

import flask
from manhattan.comparable import ComparableFrame
from manhattan.formatters.text import remove_accents
from manhattan.secure import encrypt_password, random_token
from mongoframes import ASC, In, IndexModel, Q, SortBy
import pyotp

__all__ = ['BaseUser']


class BaseUser(ComparableFrame):
    """
    The `BaseUser` collection provides user access / control to manhattan
    manage applications.

    NOTE: This is a **base** class and should not be used directly, instead
    define a new `User` class for your project that inherits from this class.
    """

    _fields = {

        # The name of the user
        'first_name',
        'last_name',

        # A lowercase version of the user's full name (for searching)
        'full_name_lower',

        # The user's email address
        'email',

        # A lowercase version of the user's email (for lookup and indexing)
        'email_lower',

        # Flag indicating the date/time a user was last sent an invite to the
        # application.
        'invited',

        # A unique token assigned to the user when invited, the token is used
        # to identify the user (as a parameter) to accept their invite.
        'invite_token',

        # Flag indicating the date/time a user accepted their invite
        'invite_accepted',

        # A hashed version of the user's password used in combination with the
        # user's salt to verify a password when supplied by the user.
        'password_hash',

        # A random string used
        'password_salt',

        # A flag indicating the date/time a user requested a reset password
        # link.
        'password_reset_requested',

        # A unique token assigned to the user when they request a password
        # reset link, the token is used to identify the user (as a parameter)
        # to set a new password.
        'password_reset_token',

        # The session token identifying the user to the manage application
        # (the session token is generated when the user signs-in and stored
        # in the secure session cookie).
        'session_token',

        # MFA (Multi-factor authentication) fields

        # Flag indicating if MFA has been enabled for this user
        'mfa_enabled',

        # The secret used to generate and verify MFA time-based one time
        # passwords (TOTP) for this user. This key is registered with an
        # authentication device (such as a mobile phone app) and then used by
        # both the app and the device to generate a TOTP.
        'mfa_otp_secret',

        # A list of emergency one time passwords for the user created when
        # multi-factor authentication is enabled.
        'mfa_recovery_codes'
    }

    _private_fields = ComparableFrame._private_fields | {
        'email_lower',
        'full_name_lower',
        'invited',
        'invite_accepted',
        'invite_token',
        'mfa_recovery_codes',
        'mfa_otp_secret',
        'password_hash',
        'password_reset_requested',
        'password_reset_token',
        'password_salt',
        'session_token'
    }

    _uncompared_fields = ComparableFrame._uncompared_fields | {
        'email_lower',
        'full_name_lower',
        'invited',
        'invite_accepted',
        'invite_token',
        'mfa_recovery_codes',
        'mfa_otp_secret',
        'password_reset_requested',
        'password_reset_token',
        'password_salt',
        'session_token'
    }

    _recommended_indexes = [
        IndexModel([('email_lower', ASC)], unique=True),
        IndexModel([('invite_token', ASC)], unique=True, sparse=True),
        IndexModel(
            [('password_reset_token', ASC)],
            unique=True,
            sparse=True
            ),
        IndexModel([('session_token', ASC)], unique=True, sparse=True)
    ]

    def __str__(self):
        return f'{self.full_name} <{self.email}>'

    # Properties

    @property
    def current_session(self):
        return self.get_session_cls().get_current_session(self._id)

    @property
    def full_name(self):
        return f'{self.first_name} {self.last_name}'

    @property
    def last_accessed(self):
        user_session = self.get_session_cls().one(
            Q.user == self._id,
            sort=SortBy(Q.last_accessed.desc),
            projection={'last_accessed': True}
        )
        if user_session:
            return user_session.last_accessed

    @property
    def mfa_authorized(self):
        current_session = self.get_session_cls().get_current_session(
            self._id,
            projection={'mfa_authorized': True}
        )
        return current_session and current_session.mfa_authorized is not None

    @property
    def password(self):
        # The 'password' property however be used to retrieve a password only
        # to set one.
        return ''

    @password.setter
    def password(self, value):
        # Requests to set the password to a none value (e.g '', or None) will
        # be ignored.
        if value in [None, '']:
            return

        # Create a random salt for the password
        self.password_salt = secrets.token_urlsafe(64)

        # Set the password
        prefix = self.get_settings_prefix()
        iterations = flask\
            .current_app\
            .config[f'{prefix}_PASSWORD_HASH_ITERATIONS']
        self.password_hash = encrypt_password(
            value,
            self.password_salt,
            iterations=iterations
        )

        # Set the user as having accepted an invite
        if not self.invited:
            self.invited = datetime.now()
        self.invite_accepted = True

    # Methods

    def generate_mfa_otp_provisioning_uri(self):
        """
        Generate a URI for provisioning multi-factor autentication for this
        user (this URL is used to generate a QR code that can be scanned by
        an authentication app such as Google authenticator).
        """
        prefix = self.get_settings_prefix()
        return pyotp.TOTP(self.mfa_otp_secret).provisioning_uri(
            self.email,
            issuer_name=current_app.config[f'{prefix}_MFA_ISSUER']
        )

    def mfa_topt_eq(self, password):
        """Verify a time-based one time password for the user"""
        return pyotp.TOTP(self.mfa_otp_secret).verify(password)

    def password_eq(self, password):
        """Check if a password is equal to the user's password"""
        prefix = self.get_settings_prefix()
        iterations = flask\
            .current_app\
            .config[f'{prefix}_PASSWORD_HASH_ITERATIONS']
        password_hash = encrypt_password(
            password,
            self.password_salt,
            iterations=iterations
        )
        return self.password_hash == password_hash

    def sign_in(self, force_new_token=False):
        """
        Sign a user in. The function returns the user session and a flag
        indicating if a new session (device) was created (True) or an existing
        known session (device) was used (False).
        """

        # Sign the user in against a session
        user_session, new_session = self.get_session_cls().sign_in(self._id)

        # Re-use an existing session token if available and not expired
        last_accessed = self.last_accessed
        if last_accessed:

            age = datetime.utcnow() - last_accessed
            prefix = self.get_settings_prefix()

            if age > flask.current_app.config[f'{prefix}_SESSION_LIFESPAN']:

                # The session token has expired remove it so a new one is
                # generated.
                self.session_token = None

        if not self.session_token or force_new_token:

            # No existing session token so create a new one
            self.session_token = random_token()

        self.update('session_token')

        return (user_session, new_session)

    # Class methods

    # Sessions management

    @classmethod
    def from_session(cls, projection=None):
        """Return a user from the current session token"""

        # Do we have session token?
        session_token = flask.session.get(cls.get_session_token_key())
        if not session_token:
            return None

        # Is the session token valid?
        user = cls.one(
            Q.session_token == session_token,
            projection=projection
        )
        if not user:
            return None

        # Verify the session/device
        prefix = cls.get_settings_prefix()
        lifespan = flask.current_app.config[f'{prefix}_SESSION_LIFESPAN']
        if not cls.get_session_cls().verify_access(user._id, lifespan):
            return None

        return user

    @classmethod
    def get_g_key(cls):
        """
        Return the key used to store an instance of the user against the
        global context (`flask.g`).
        """
        return 'user'

    @classmethod
    def get_session_cls(cls):
        """Return the session class that will be used by the user"""
        raise NotImplemented()

    @classmethod
    def get_session_token_key(cls):
        """
        Get the session token key used to store/retrieve the user's session
        token in the flask session.
        """
        return 'manage_session_token'

    @classmethod
    def get_settings_prefix(cls):
        """Return the prefix for settings properties for the user"""
        return 'USER'

    # Failed sign-ins

    @classmethod
    def get_sign_in_attempt_key(cls, email):
        """
        Return a key that can be used to store and retrieve the number of
        failed sign-in attempts for a given email against this application.
        """
        server_name = flask.current_app.config['SERVER_NAME']
        return f'user_failed_login:{server_name}:{email.lower()}'

    @classmethod
    def is_locked_out(cls, email):
        """
        Return True if the given email is locked out (cannot be used to
        sign-in until the lock is removed or expires.)
        """
        config = flask.current_app.config
        prefix = cls.get_settings_prefix()
        cache = config[f'{prefix}_FAILED_SIGN_IN_CACHE']
        return (cache.get(cls.get_sign_in_attempt_key(email)) or 0) \
                >= config[f'{prefix}_MAX_FAILED_SIGN_IN_ATTEMPTS']

    # Static methods

    @staticmethod
    def _on_upsert(sender, frames):
        for frame in frames:

            if frame.email:

                # Store lowercase email
                frame.email_lower = frame.email.lower()

            if frame.first_name and frame.last_name:

                # Store lowercase full name
                frame.full_name_lower = frame.full_name.replace(',','').lower()
                frame.full_name_lower = remove_accents(frame.full_name_lower)

    @staticmethod
    def _on_delete(sender, frames):

        # Delete any associated user sessions
        user_cls = sender.get_session_cls()
        user_cls.delete_many(
            user_cls.many(In(Q.user, [f._id for f in frames]))
        )




"""
Revoke a user's session.
"""

import bson

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.manage.views import factories


__all__ = ['revoke_session_chains']


# Chains

revoke_session_chains = ChainMgr()

# POST
revoke_session_chains['post'] = Chain([
    'config',
    'authenticate',
    'get_session',
    'revoke_session',
    'redirect'
])

# Define the links
revoke_session_chains.set_link(factories.config())
revoke_session_chains.set_link(factories.authenticate())

@revoke_session_chains.link
def get_session(state):
    """Get the user's session"""

    user_cls = state.manage_config.frame_cls
    user_session_cls = user_cls.get_session_cls()

    # Attempt to convert the session Id to the required type
    user_session_id = flask.request.values.get('session')
    try:
        user_session_id = bson.objectid.ObjectId(user_session_id)
    except bson.errors.InvalidId:
        abort(404)

    # Attempt to retrieve the document
    state.user_session = user_session_cls.by_id(user_session_id)
    if not state.user_session:
        abort(404)

    # Find the user associated with the session
    user = user_cls.by_id(state.user_session.user)
    state[state.manage_config.var_name] = user

    # If the endpoint is `revoke_my_session` check that the user removing the
    # session is the one who's currently signed in.
    if flask.request.endpoint == 'revoke_my_session':
        if user != getattr(flask.g, user_cls.get_g_key()):
            flask.abort(403)

@revoke_session_chains.link
def revoke_session(state):
    """Revoke the session for the user"""

    # Remove the session
    state.user_session.delete()

    # Notify the user that the session has been revoked
    flask.flash('Session revoked')

@revoke_session_chains.link(name='redirect')
def _redirect(state):
    """Redirect the user depending on the endpoint used to call the view"""
    user = state[state.manage_config.var_name]

    if flask.request.endpoint.endswith('revoke_my_session'):
        return flask.redirect(
            flask.url_for(state.manage_config.get_endpoint('security'))
        )

    else:
        return flask.redirect(
            flask.url_for(
                state.manage_config.get_endpoint('view'),
                **{state.manage_config.var_name: user._id}
            )
        )



"""
Allow users to accept an invite to join the application.
"""

from datetime import datetime

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.comparable.change_log import ChangeLogEntry
from manhattan.forms import BaseForm, fields, validators
from manhattan.manage.views import factories, utils
from manhattan.users.utils import apply_password_rules, password_validator
from mongoframes import Q

__all__ = ['accept_invite_chains']


# Forms

class AcceptInviteForm(BaseForm):

    token = fields.HiddenField()

    password = fields.PasswordField(
        'Password',
        [
            validators.Required(),
            validators.EqualTo('confirm_password', "Passwords don't match."),
            password_validator
        ]
    )

    confirm_password = fields.PasswordField('Confirm password')


# Chains
accept_invite_chains = ChainMgr()

# GET
accept_invite_chains['get'] = Chain([
    'config',
    'init_form',
    'apply_password_rules',
    'get_user',
    'decorate',
    'render_template'
])

# POST
accept_invite_chains['post'] = Chain([
    'config',
    'init_form',
    'apply_password_rules',
    'get_user',
    'validate',
    [
        [
            'accept_invite',
            'sign_in',
            'redirect'
        ], [
            'decorate',
            'render_template'
        ]
    ]
])

# Define the links
accept_invite_chains.set_link(factories.config(form_cls=AcceptInviteForm))
accept_invite_chains.set_link(factories.init_form())
accept_invite_chains.set_link(apply_password_rules())
accept_invite_chains.set_link(factories.validate())
accept_invite_chains.set_link(factories.render_template('accept_invite.html'))
accept_invite_chains.set_link(factories.redirect('dashboard'))

@accept_invite_chains.link
def decorate(state):
    state.decor = utils.base_decor(state.manage_config, 'accept_invite')
    state.decor['title'] = 'Accept invite'

@accept_invite_chains.link
def get_user(state):
    """Get the user who is accepting an invite base on the token"""
    user_cls = state.manage_config.frame_cls
    prefix = user_cls.get_settings_prefix()

    request_new_invite_url = flask.url_for(
        state.manage_config.get_endpoint('request_new_invite')
    )

    invite_token = flask.request.values.get('invite_token', '').strip()
    if not invite_token:

        # The token was provided, notify the user and redirect them to request
        # a new invite.
        flask.flash(
            'Not a valid invite link, no invite token provided.',
            'warning'
        )
        return flask.redirect(request_new_invite_url)

    # Get the user by their invite token
    user = user_cls.one(Q.invite_token == invite_token)
    state[state.manage_config.var_name] = user

    if not user:

        # No user found for the invite token, notify the user and redirect
        # them to request a new invite.
        flask.flash(
            'Not a valid invite link, the invite token does not exist.',
            'warning'
        )
        return flask.redirect(request_new_invite_url)

    ellapsed = datetime.utcnow() - user.invited
    if ellapsed > flask.current_app.config.get(f'{prefix}_INVITE_LIFESPAN'):

        # The invite has expired, notify the user and redirect them to request
        # a new invite.
        flask.flash(
            'It appears this invite has expired please request a new one.',
            'warning'
        )
        return flask.redirect(request_new_invite_url)

    if user_cls.is_locked_out(user.email):

        # The user is currently locked out for too many failed attempts to
        # sign-in attempts, notify the user and redirect them to the accept
        # invite page.
        lockout = flask.current_app.config[f'{prefix}_FAILED_SIGN_IN_LOCKOUT']
        minutes = round(lockout.total_seconds() / 60)
        flash(
            f'Your account is locked out due to too many failed sign '
            f'in attempts, please try again in {minutes} minutes.',
            'error'
        )
        return flask.redirect(flask.request.full_path)

@accept_invite_chains.link
def accept_invite(state):
    """Accept an invite for a user"""
    user = state[state.manage_config.var_name]

    # Set the user's password
    user.password = state.form.data['password']
    user.update(
        'invite_accepted',
        'invited',
        'modified',
        'password_hash',
        'password_salt'
    )

    # Log the change
    entry = ChangeLogEntry({
        'type': 'NOTE',
        'documents': [user],
        'user': user
    })
    entry.add_note('Accepted invite')
    entry.insert()

    # Clear the invite token so the invite can't be used again
    state.manage_config.frame_cls.get_collection().update(
        {'_id': user._id},
        {'$unset': {'invite_token': ''}}
    )

@accept_invite_chains.link
def sign_in(state):
    """Sign the user in after they have accepted their invite"""
    user_cls = state.manage_config.frame_cls
    user = state[state.manage_config.var_name]

    user.sign_in()
    flask.session[user_cls.get_session_token_key()] = user.session_token



"""
View a user's activity log (the change log entries they are responsible for).
"""

from datetime import datetime, time
from urllib.parse import urlencode

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.comparable.change_log import ChangeLogEntry
from manhattan.forms import BaseForm, fields, validators
from manhattan.manage.views import factories, generic, utils
from manhattan.nav import Nav, NavItem
from mongoframes import And, InvalidPage, Paginator, Q, SortBy

__all__ = ['activity_log_chains']


# Forms


class Filters(BaseForm):

    _inline = True

    from_date = fields.DateField(
        'From',
        render_kw={
            'data-mh-date-picker': True,
            'placeholder': '1 Jan 2000'
        }
    )

    to_date = fields.DateField(
        'To',
        render_kw={
            'data-mh-date-picker': True,
            'placeholder': '1 Jan 2000'
        }
    )

    def filter_from_date(filter_form, form, field):
        if field.data:
            return Q.created >= datetime.combine(field.data, time.min)

    def filter_to_date(filter_form, form, field):
        if field.data:
            return Q.created <= datetime.combine(field.data, time.max)


class ListForm(BaseForm):

    _info = {'show_search_button': True}

    page = fields.IntegerField('Page', default=1)

    filters = fields.FormField(Filters)


# Chains
activity_log_chains = ChainMgr()

# GET
activity_log_chains['get'] = Chain([
    'config',
    'authenticate',
    'get_document',
    'init_form',
    'validate',
    'filter',
    'related_filter',
    'paginate',
    'form_info',
    'decorate',
    'render_template'
])

# Define the links
activity_log_chains.set_link(
    factories.config(
        change_log_cls=ChangeLogEntry,
        collation=None,
        form_cls=ListForm,
        hint=None,
        orphans=2,
        per_page=20,
        projection=None
    )
)
activity_log_chains.set_link(factories.authenticate())
activity_log_chains.set_link(factories.get_document())
activity_log_chains.set_link(generic.list['get'].get_link('validate'))
activity_log_chains.set_link(generic.list['get'].get_link('filter'))
activity_log_chains.set_link(generic.list['get'].get_link('form_info'))
activity_log_chains.set_link(factories.render_template('activity_log.html'))

@activity_log_chains.link
def decorate(state):

    document = state[state.manage_config.var_name]

    state.decor = utils.base_decor(
        state.manage_config,
        state.view_type,
        document
    )

    # Title
    state.decor['title'] = state.manage_config.titleize(document)

    # Breadcrumbs
    if Nav.exists(state.manage_config.get_endpoint('list')):
        state.decor['breadcrumbs'].add(
            utils.create_breadcrumb(state.manage_config, 'list')
        )

    if Nav.exists(state.manage_config.get_endpoint('view')):
        state.decor['breadcrumbs'].add(
            utils.create_breadcrumb(state.manage_config, 'view', document)
        )

    state.decor['breadcrumbs'].add(NavItem('Activity log'))

@activity_log_chains.link
def init_form(state):
    # Ensure the form class has a hidden field to indicate which document the
    # change log should relate to.

    if hasattr(state.form_cls, state.manage_config.var_name):
        form_cls = state.form_cls

    else:
        class form_cls(state.form_cls):
            pass

        setattr(form_cls, state.manage_config.var_name, fields.HiddenField())

    # Initialize the form
    state.form = form_cls(
        flask.request.args or None
    )

@activity_log_chains.link
def related_filter(state):
    document = state[state.manage_config.var_name]

    # Filter the results by which user made the change
    if state.query:
        state.query = And(state.query, Q.user == document._id)
    else:
        state.query = Q.user == document._id

@activity_log_chains.link
def paginate(state):

    # Select the documents in paginated form
    paginator_kw = {
        'per_page': state.per_page,
        'orphans': state.orphans,
        'sort': SortBy(Q.created.desc)
    }

    if state.projection :
        paginator_kw['projection'] = state.projection

    if state.collation:
        paginator_kw['collation'] = state.collation

    if state.hint:
        paginator_kw['hint'] = state.hint

    # Select the documents in paginated form
    state.paginator = Paginator(
        state.change_log_cls,
        state.query or {},
        **paginator_kw
        )

    # Select the requested page
    try:
        state.page = state.paginator[state.form.data.get('page', 1)]
    except InvalidPage:
        return flask.redirect(flask.url_for(flask.request.url_rule.endpoint))



import flask
from manhattan.forms import validators

__all__ = [

    # Factories
    'apply_password_rules',

    # Validators
    'password_validator'
]


# Factories

def apply_password_rules():
    """
    Apply the configured password rules to the form so that any
    `password_validator` will use them.
    """

    def apply_password_rules(state):
        user_cls = state.manage_config.frame_cls
        prefix = user_cls.get_settings_prefix()

        # Store the password rules against the form
        state.form._password_rules = flask\
            .current_app\
            .config[f'{prefix}_PASSWORD_RULES']

    return apply_password_rules

# Validators

def password_validator(form, field):
    """
    Wrapper for the password validator so that rules for passwords can be
    defined in the applications settings (config). See `password_validator`.
    """
    return validators.Password(**form._password_rules)(form, field)



"""
Sign the current user out.
"""

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.manage.views import factories

__all__ = ['sign_out_chains']


# Chains
sign_out_chains = ChainMgr()

# GET
sign_out_chains['get'] = Chain([
    'config',
    'sign_out',
    'redirect'
    ])

# Define the links
sign_out_chains.set_link(factories.config())
sign_out_chains.set_link(factories.redirect('sign_in'))

@sign_out_chains.link
def sign_out(state):
    user_cls = state.manage_config.frame_cls
    user = getattr(flask.g, user_cls.get_g_key(), None)

    # Clear the user's session token
    if user:
        user_cls.get_collection().update(
            {'_id': user._id},
            {'$unset': {'session_token': ''}}
        )

    # Clear the session token from the the flask session
    if user_cls.get_session_token_key() in flask.session:
        flask.session.pop(user_cls.get_session_token_key())



"""
Authorize the user using a secondary mechanism.
"""

from datetime import datetime
import io
import os
import urllib.parse as urlparse
from urllib.parse import urlencode

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.forms import BaseForm, fields, validators
from manhattan.manage.views import factories, utils
from manhattan.nav import NavItem
from manhattan import secure
import pyotp

__all__ = ['auth_chains']


# Utils

def is_blocked(user):
    """
    Return true if the user is temporarily blocked from authorizing due to too
    many failed attempts.
    """
    prefix = user.get_settings_prefix()
    cache = flask.current_app.config[f'{prefix}_MFA_FAILED_AUTH_CACHE']
    key = f'mfa_failed_auth:{user._id}'

    max_attempts = flask\
        .current_app\
        .config[f'{prefix}_MFA_MAX_FAILED_AUTH_ATTEMPTS']

    return (cache.get(key) or 0) >= max_attempts


# Forms

class AuthForm(BaseForm):

    totp = fields.StringField(
        'Your six-digit code',
        [validators.Required()],
        render_kw={'autofocus': True}
    )

    def validate_totp(form, field):
        if not field.data:
            return

        # Check that the user has not been blocked due to too many failed
        # attempts.
        if is_blocked(form.obj):
            raise validators.ValidationError('To many failed attempts.')

        # Check that the code given is valid
        verified = False

        # First check the TOTP...
        topt = field.data.replace(' ', '')
        if pyotp.TOTP(form.obj.mfa_otp_secret).verify(topt):
            verified = True

        # ...then check the recovery codes for the user
        elif topt in form.obj.mfa_recovery_codes:
            verified = True

        if not verified:
            raise validators.ValidationError('Not a valid code')


# Chains
auth_chains = ChainMgr()

# POST
auth_chains['get'] = Chain([
    'config',
    'authenticate',
    'init_form',
    'decorate',
    'render_template'
])

# POST
auth_chains['post'] = Chain([
    'config',
    'authenticate',
    'init_form',
    'validate',
    [
        ['authorize'],
        [
            'log_failed_attempts',
            'decorate',
            'render_template'
        ]
    ]
])

# Define the links
auth_chains.set_link(
    factories.config(
        form_cls=AuthForm,
        open_user_nav=True
    )
)
auth_chains.set_link(factories.validate())
auth_chains.set_link(factories.render_template('mfa/auth.html'))

@auth_chains.link
def authenticate(state):
    """Ensure there's a signed in user"""
    user_cls = state.manage_config.frame_cls
    if not getattr(flask.g, user_cls.get_g_key(), None):
        return flask.redirect(
            flask.url_for(state.manage_config.get_endpoint('sign_in'))
        )

@auth_chains.link
def decorate(state):
    user_cls = state.manage_config.frame_cls
    prefix = user_cls.get_settings_prefix()

    state.decor = utils.base_decor(state.manage_config, 'mfa_auth')
    state.decor['title'] = '2-factor authentication'
    state.decor['breadcrumbs'].add(NavItem('2-factor authentication', ''))

    state.mfa_issuer = flask.current_app.config[f'{prefix}_MFA_ISSUER']

@auth_chains.link
def init_form(state):
    user_cls = state.manage_config.frame_cls

    # Initialize the form
    state.form = state.form_cls(
        flask.request.form,
        obj=getattr(flask.g, user_cls.get_g_key())
    )

    # Store the user class against the form
    state.form._user_cls = user_cls

@auth_chains.link
def log_failed_attempts(state):
    """Log a failed attempt to authorize using MFA"""

    if not state.form.totp.data:
        # Don't log failed attempts if no password (totp) was provided
        return

    user_cls = state.manage_config.frame_cls
    prefix = user_cls.get_settings_prefix()
    user = getattr(flask.g, user_cls.get_g_key())
    cache = flask.current_app.config[f'{prefix}_MFA_FAILED_AUTH_CACHE']
    attempt_key = f'mfa_failed_auth:{user._id}'
    attempts = cache.get(attempt_key) or 0
    lockout = flask.current_app.config[f'{prefix}_MFA_FAILED_AUTH_LOCKOUT']

    # Record the failed attempt
    cache.set(attempt_key, attempts + 1, lockout)

    if is_blocked(user):

        # Notify the user that they are temporarily blocked from
        # authorizing.
        flask.session['_flashes'] = []
        seconds = round(lockout.total_seconds())
        flask.flash(
            f'To too many failed attempts. Please wait {seconds} seconds '
            'before trying again.',
            'error'
        )

@auth_chains.link
def authorize(state):
    """
    Create a session key which can be used by the caller to authorize the
    request and return the user to the caller with the key.
    """
    user_cls = state.manage_config.frame_cls
    user = getattr(flask.g, user_cls.get_g_key())

    # Check if a recovery code was used
    totp = state.form.data['totp']
    if totp in user.mfa_recovery_codes:

        # Remove the recovery code for the user
        user_cls.get_collection().update(
            {'_id': user._id},
            {'$pull': {'mfa_recovery_codes': totp}}
        )

        # Email the user to let them know a recovery code was used
        env = flask.current_app.jinja_env
        template_path = os.path.join(
            state.manage_config.template_path,
            'emails/recovery_code_used.html'
        )

        flask.current_app.manage.send_email(
            [user.email],
            'Recovery code used',
            template_path,
            global_vars={'user': user.to_json_type()},
            template_map={
                'recovery_code_used': env.loader.get_source(
                    env,
                    'manhattan/users/emails/recovery_code_used.html'
                )[0],
                template_path: env.loader.get_source(env, template_path)[0]
            }
        )

    # Verify the caller URL is safe
    caller_url = flask.request.args.get('caller_url')
    caller_url = secure.safe_redirect_url(caller_url) if caller_url else None

    if caller_url is None:
        flask.flash('Invalid caller URL', 'error')
        return flask.redirect(
            flask.url_for(state.manage_config.get_endpoint('dashboard'))
        )

    # Update the user's sessions MFA authorized date/time
    flask.current_session = user.current_session
    flask.current_session.mfa_authorized = datetime.utcnow()
    flask.current_session.update('mfa_authorized')

    # Check to see if the caller is requesting a scoped session
    if 'scoped' in flask.request.args:

        # Generate the scoped session key
        prefix = user_cls.get_settings_prefix()
        session_token = secure.mfa.create_scoped_session(
            flask.current_app.config[f'{prefix}_MFA_SCOPED_SESSION_CACHE'],
            (str(user._id), caller_url),
            flask.current_app.config[f'{prefix}_MFA_SCOPED_SESSION_LIFESPAN']
        )

        # Add the scoped session key to the caller URL

        # Parse the URL
        caller_url_parts = list(urlparse.urlparse(caller_url))

        # Update the query parameters
        query = dict(urlparse.parse_qsl(caller_url_parts[4]))
        query['mfa_scoped_session_token'] = session_token
        caller_url_parts[4] = urlencode(query)

        # Rebuild the URL
        caller_url = urlparse.urlunparse(caller_url_parts)

    # Return the user to the caller
    return flask.redirect(caller_url)



"""
Regenerate the user's list of recovery codes.
"""

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.manage.views import factories
from manhattan.secure import random_code

__all__ = ['regenerate_recovery_codes_chains']


# Define the chains
regenerate_recovery_codes_chains = ChainMgr()

# POST
regenerate_recovery_codes_chains['post'] = Chain([
    'config',
    'authenticate',
    'regenerate_recovery_codes',
    'redirect'
])

regenerate_recovery_codes_chains.set_link(factories.config())
regenerate_recovery_codes_chains.set_link(factories.authenticate())
regenerate_recovery_codes_chains.set_link(
    factories.redirect('mfa_recovery_codes')
)

@regenerate_recovery_codes_chains.link
def regenerate_recovery_codes(state):
    user_cls = state.manage_config.frame_cls
    prefix = user_cls.get_settings_prefix()

    # Generate the new recovery codes
    recovery_codes = []
    recovery_code_count = flask\
        .current_app\
        .config[f'{prefix}_MFA_RECOVERY_CODES']

    for i in range(0, recovery_code_count):
        recovery_codes.append(
            random_code(
                flask.current_app.config[f'{prefix}_MFA_RECOVERY_CODE_LENGTH']
            )
        )

    # Save the new recovery codes
    state.manage_user.logged_update(
        state.manage_user,
        {'mfa_recovery_codes': recovery_codes},
        'mfa_recovery_codes',
        'modified'
    )

    # Notify the user that the recovery codes have been regenerated
    flask.flash('New recovery codes generated')



"""
Resend an invite to a user.
"""

from datetime import datetime
import os

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.formatters.flask import path_to_url
from manhattan.manage.views import factories
from manhattan.secure import random_token

__all__ = ['resend_invite_chains']


# Chains

resend_invite_chains = ChainMgr()

# POST
resend_invite_chains['post'] = Chain([
    'authenticate',
    'get_document',
    'resend_invite',
    'redirect'
])

# Define the links
resend_invite_chains.set_link(factories.authenticate())
resend_invite_chains.set_link(factories.get_document())
resend_invite_chains.set_link(factories.redirect('view', include_id=True))

@resend_invite_chains.link
def resend_invite(state):
    """Resend the user an invite"""
    env = flask.current_app.jinja_env
    template_path = os.path.join(
        state.manage_config.template_path,
        'emails/invite.html'
    )
    user = state[state.manage_config.var_name]

    # Generate the invite for the user
    user.invited = datetime.utcnow()
    user.invite_token = random_token()
    user.update('invited', 'invite_token')

    # Send the user an invite by email
    flask.current_app.manage.send_email(
        [user.email],
        'Your invite',
        template_path,
        global_vars={
            'user': user.to_json_type(),
            'invite_link': path_to_url(
                flask.url_for(
                    state.manage_config.get_endpoint('accept_invite'),
                    invite_token=user.invite_token
                )
            )
        },
        template_map={
            'invite': env.loader.get_source(
                env,
                'manhattan/users/emails/invite.html'
            )[0],
            template_path: env.loader.get_source(env, template_path)[0]
        }
    )

    # Notify the user that a new invite has been sent
    flask.flash('New invite sent.')



"""
List users.
"""

from manhattan.forms import BaseForm, fields
from manhattan.forms.utils import sort_by_choices
from manhattan.manage.views import factories, generic

__all__ = ['list_chains']


# Forms

class ListForm(BaseForm):

    page = fields.IntegerField('Page', default=1)

    q = fields.StringField('Search')

    sort_by = fields.SelectField(
        'Sort by',
        choices=sort_by_choices([
            ('full_name_lower', 'Name')
        ]),
        default='full_name_lower'
    )


# Chains
list_chains = generic.list.copy()

# Factory overrides
list_chains.set_link(
    factories.config(
        collation=None,
        form_cls=ListForm,
        hint=None,
        orphans=2,
        per_page=20,
        projection={
            'first_name': True,
            'last_name': True
        },
        search_fields=['full_name_lower']
    )
)



from datetime import timedelta
from manhattan.utils.cache import MemoryCache


__all__ = [

    # Factories
    'createDefaultConfig',

    # Classes
    'DefaultConfig'
]


def createDefaultConfig(prefix='USER'):
    """Return a default config mixin class for the blueprint"""

    class DefaultConfig:
        """
        Default settings for the the user access / control set up.
        """

    # Invites

    # The time before an invite to a user to join the application expires
    setattr(
        DefaultConfig,
        f'{prefix}_INVITE_LIFESPAN',
        timedelta(days=14)
    )

    # Sessions

    # The maximum number of sessions (devices) a user can have registered at
    # any one time.
    setattr(DefaultConfig, f'{prefix}_MAX_SESSIONS', 12)

    # The time before a user's session expires (from the last time the user
    # accessed the application).
    setattr(
        DefaultConfig,
        f'{prefix}_SESSION_LIFESPAN',
        timedelta(days=14)
    )


    # Passwords

    # The number of iterations used when hashing a user's password. You might
    # want to configure this to a low number in local development environments
    # to ensure generating users from fixtures doesn't take too long. We don't
    # recommend less that 100,000 in production.
    setattr(
        DefaultConfig,
        f'{prefix}_PASSWORD_HASH_ITERATIONS',
        100000
    )

    # The time before a reset password link for a user expires
    setattr(
        DefaultConfig,
        f'{prefix}_PASSWORD_RESET_LIFESPAN',
        timedelta(days=14)
    )

    # The set of rules for users when setting a password
    setattr(
        DefaultConfig,
        f'{prefix}_PASSWORD_RULES',
        dict(
            min_length=10,
            max_length=128,
            min_lower=1,
            min_upper=1,
            min_digits=1,
            min_specials=1
        )
    )


    # Failed sign-in attempts and lockouts

    # The cache used to store the number of failed attempts to sign-in for a
    # user identified (e.g an email).
    setattr(
        DefaultConfig,
        f'{prefix}_FAILED_SIGN_IN_CACHE',
        MemoryCache()
    )

    # The period of time a user is locked out for after too many failed
    # sign-in attempts.
    setattr(
        DefaultConfig,
        f'{prefix}_FAILED_SIGN_IN_LOCKOUT',
        timedelta(minutes=30)
    )

    # The window of time during which failed attempts are tallied before the
    # the tally is reset (the user can make `USER_MAX_FAILED_SIGN_IN_ATTEMPTS`
    # number of failed attempts within this window).
    setattr(
        DefaultConfig,
        f'{prefix}_FAILED_SIGN_IN_WINDOW',
        timedelta(minutes=10)
    )

    # The maximum number of times a user can fail to sign-in before being
    # locked out.
    setattr(
        DefaultConfig,
        f'{prefix}_MAX_FAILED_SIGN_IN_ATTEMPTS',
        3
    )


    # MFA (Multi-factor authorization)

    # The cache used to store the number of failed attempts to authorize using
    # multi-factor authentication.
    setattr(
        DefaultConfig,
        f'{prefix}_MFA_FAILED_AUTH_CACHE',
        MemoryCache()
    )

    # The duration a user is locked out of MFA after failing to authorize,
    # this should be a short period (less than 30 seconds) set to throttle
    # the number of attempts that can be made for the time-based one time
    # password (TOTP).
    setattr(
        DefaultConfig,
        f'{prefix}_MFA_FAILED_AUTH_LOCKOUT',
        timedelta(seconds=10)
    )

    # The MFA provisioning issuer
    setattr(DefaultConfig, f'{prefix}_MFA_ISSUER', 'NOT SET')

    # The number of times a user can incorrectly enter their time-based one
    # time password (TOPT) before we temporarily block their account from MFA.
    setattr(
        DefaultConfig,
        f'{prefix}_MFA_MAX_FAILED_AUTH_ATTEMPTS',
        3
    )

    # The length of MFA recovery codes generated for a user
    setattr(
        DefaultConfig,
        f'{prefix}_MFA_RECOVERY_CODE_LENGTH',
        10
    )

    # The number of MFA recovery codes to generate for a user
    setattr(
        DefaultConfig,
        f'{prefix}_MFA_RECOVERY_CODES',
        16
    )

    # Flag indicating if MFA is required for users
    setattr(DefaultConfig, f'{prefix}_MFA_REQUIRED', False)

    # The cache used to store scoped session tokens
    setattr(
        DefaultConfig,
        f'{prefix}_MFA_SCOPED_SESSION_CACHE',
        MemoryCache()
    )

    # The period of time a scoped authorized session lasts (we recommended
    # under 30 minutes).
    setattr(
        DefaultConfig,
        f'{prefix}_MFA_SCOPED_SESSION_LIFESPAN',
        timedelta(minutes=10)
    )

    return DefaultConfig


DefaultConfig = createDefaultConfig()



"""
Allow users to request a link to reset their password with.
"""

from datetime import datetime
import os

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.formatters.flask import path_to_url
from manhattan.forms import BaseForm, fields, validators
from manhattan.manage.views import factories, utils
from manhattan.secure import random_token
from mongoframes import Q

__all__ = ['reset_password_chains']


# Forms

class ResetPasswordForm(BaseForm):

    email = fields.StringField(
        'Email',
        [validators.Required(), validators.Email()]
    )


# Chains
reset_password_chains = ChainMgr()

# GET
reset_password_chains['get'] = Chain([
    'config',
    'init_form',
    'decorate',
    'render_template'
])

# POST
reset_password_chains['post'] = Chain([
    'config',
    'init_form',
    'validate',
    [
        [
            'get_user',
            'limit_rate',
            'send_reset_link',
            'redirect'
        ],
        [
            'decorate',
            'render_template'
        ]
    ]
])

# Define the links
reset_password_chains.set_link(factories.config(form_cls=ResetPasswordForm))
reset_password_chains.set_link(factories.validate())
reset_password_chains.set_link(
    factories.render_template('reset_password.html')
)
reset_password_chains.set_link(factories.redirect('reset_password'))

@reset_password_chains.link
def decorate(state):
    state.decor = utils.base_decor(state.manage_config, 'reset_password')
    state.decor['title'] = 'Reset password'

@reset_password_chains.link
def init_form(state):
    # Initialize the form
    state.form = state.form_cls(flask.request.form)

    # Store the user class against the form
    state.form._user_cls = state.manage_config.frame_cls

@reset_password_chains.link
def get_user(state):
    """Get the user to send a reset link to"""
    user_cls = state.manage_config.frame_cls
    user = user_cls.one(Q.email_lower == state.form.data['email'].lower())
    state[state.manage_config.var_name] = user

    if not user:
        # If we couldn't find a user based on the email address the user gave
        # then we don't indicate this to the user as this can be exploited to
        # determine registered email addresses.
        flask.flash(
            'Instructions on resetting your password have been emailed to you.'
        )

        # Redirect the user back to the reset password page
        return flask.redirect(
            flask.url_for(state.manage_config.get_endpoint('reset_password'))
        )

@reset_password_chains.link
def limit_rate(state):
    """Limit the number of reset requests a user can make to once a minute"""
    user = state[state.manage_config.var_name]

    if user.password_reset_requested is None:
        # No reset token set for the user no need to rate limit
        return

    last_request = user.password_reset_requested
    if (datetime.utcnow() - last_request).total_seconds() < 60:

        # Let the user know an email has been sent to them but don't send one
        # as we did only 60 seconds ago.
        #
        # NOTE: We used to flag to users that they should wait but this can be
        # exploited to determine registered email addresses and so we no
        # longer provide this feedback.
        #
        flask.flash(
            'Instructions on resetting your password have been emailed to you.'
        )

        # Redirect the user back to the reset password page
        return flask.redirect(
            flask.url_for(state.manage_config.get_endpoint('reset_password'))
        )

@reset_password_chains.link
def send_reset_link(state):
    """Send the user a reset password link"""

    env = flask.current_app.jinja_env
    template_path = os.path.join(
        state.manage_config.template_path,
        'emails/reset_password.html'
    )
    user = state[state.manage_config.var_name]

    # Generate the password reset token for the user
    user.password_reset_requested = datetime.utcnow()
    user.password_reset_token = random_token()
    user.update('password_reset_requested', 'password_reset_token')

    # Send the user a reset password link by email
    flask.current_app.manage.send_email(
        [user.email],
        'How to reset your password',
        template_path,
        global_vars={
            'user': user.to_json_type(),
            'reset_link':
                path_to_url(
                    flask.url_for(
                        state.manage_config.get_endpoint('set_new_password'),
                        reset_token=user.password_reset_token
                    )
                )
        },
        template_map={
            'reset_password': env.loader.get_source(
                env,
                'manhattan/users/emails/reset_password.html'
            )[0],
            template_path: env.loader.get_source(env, template_path)[0]
        }
    )

    # Let the user know to expect an email containing their reset password
    # instructions.
    flask.flash(
        'Instructions on resetting your password have been emailed to you.'
    )



"""
Download a text file containing the user's recovery codes.
"""

import io

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.manage.views import factories

__all__ = ['download_recovery_codes_chains']


# Chains
download_recovery_codes_chains = ChainMgr()

# POST
download_recovery_codes_chains['get'] = Chain([
    'config',
    'authenticate',
    'download_recovery_codes'
])

# Define the links
download_recovery_codes_chains.set_link(factories.config())
download_recovery_codes_chains.set_link(factories.authenticate())

@download_recovery_codes_chains.link
def download_recovery_codes(state):

    user_cls = state.manage_config.frame_cls
    user = getattr(flask.g, user_cls.get_g_key())

    f = io.BytesIO()
    f.write('\n'.join(user.mfa_recovery_codes).encode('utf8'))
    f.seek(0)

    return flask.send_file(
        f,
        mimetype='text/plain',
        as_attachment=True,
        attachment_filename='recovery_codes.txt'
    )



"""
Allow a user to update their profile.
"""

import os

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.forms import BaseForm, fields, validators
from manhattan.manage.views import factories, utils
from manhattan.nav import NavItem
from mongoframes import Q

__all__ = ['update_my_profile_chains']


# Forms

class UpdateMyProfileForm(BaseForm):

    first_name = fields.StringField(
        'First name',
        [
            validators.Required(),
            validators.Length(max=20)
        ])

    last_name = fields.StringField(
        'Last name',
        [
            validators.Required(),
            validators.Length(max=20)
        ])

    email = fields.StringField(
        'Email',
        [
            validators.Required(),
            validators.Email(),
            validators.Length(max=254)
        ])

    def validate_email(form, field):
        """Validate any new email address is unique"""
        user_cls = form._user_cls

        # The email address can stay the same for the user (irrespective of
        # case).
        if field.data.lower() == form.obj.email_lower:
            return

        user = user_cls.one(
            Q.email_lower == field.data.lower(),
            projection={'_id': True}
        )
        if user:
            raise validators.ValidationError(
                'This email address is already registered.'
            )


# Chains
update_my_profile_chains = ChainMgr()

# GET
update_my_profile_chains['get'] = Chain([
    'config',
    'authenticate',
    'init_form',
    'decorate',
    'render_template'
])

# POST
update_my_profile_chains['post'] = Chain([
    'config',
    'authenticate',
    'init_form',
    'validate',
    [
        [
            'update_profile',
            'redirect'
        ], [
            'decorate',
            'render_template'
        ]
    ]
])

# Define the links
update_my_profile_chains.set_link(
    factories.config(
        form_cls=UpdateMyProfileForm,
        open_user_nav=True
    )
)
update_my_profile_chains.set_link(factories.authenticate())
update_my_profile_chains.set_link(factories.validate())
update_my_profile_chains.set_link(
    factories.render_template('update_my_profile.html'))
update_my_profile_chains.set_link(factories.redirect('update_my_profile'))

@update_my_profile_chains.link
def decorate(state):
    state.decor = utils.base_decor(state.manage_config, 'update_my_profile')
    state.decor['title'] = 'Update my profile'
    state.decor['breadcrumbs'].add(NavItem('Update my profile', ''))

@update_my_profile_chains.link
def init_form(state):
    user_cls = state.manage_config.frame_cls

    # Initialize the form
    state.form = state.form_cls(
        flask.request.form,
        obj=getattr(flask.g, user_cls.get_g_key())
    )

    # Store the user class against the form
    state.form._user_cls = user_cls

@update_my_profile_chains.link
def update_profile(state):
    user_cls = state.manage_config.frame_cls
    user = getattr(flask.g, user_cls.get_g_key())

    # Update the user's profile (capture the email before hand to determine if
    # it's been changed).
    original_email = user.email
    user.logged_update(user, state.form.data)

    if original_email.lower() != user.email.lower():

        env = flask.current_app.jinja_env
        template_path = os.path.join(
            state.manage_config.template_path,
            'emails/email_changed.html'
        )

        # The user's email address has changed email both the original and new
        # email addresses notifying the user of the change.
        flask.current_app.manage.send_email(
            [user.email, original_email],
            'Your email address has been changed',
            template_path,
            global_vars={
                'user': user.to_json_type(),
                'original_email': original_email
            },
            template_map={
                'email_changed': env.loader.get_source(
                    env,
                    'manhattan/users/emails/email_changed.html'
                )[0],
                template_path: env.loader.get_source(env, template_path)[0]
            }
        )

    # Notify the user that their profile has been updated
    flask.flash('Profile updates made.')



"""
Allow users to set a new password.
"""

from datetime import datetime

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.comparable.change_log import ChangeLogEntry
from manhattan.forms import BaseForm, fields, validators
from manhattan.manage.views import factories, utils
from manhattan.users.utils import apply_password_rules, password_validator
from mongoframes import Q

__all__ = ['set_new_password_chains']


# Forms

class SetNewPasswordForm(BaseForm):

    reset_token = fields.HiddenField()

    password = fields.PasswordField(
        'Password',
        [
            validators.Required(),
            password_validator,
            validators.EqualTo('confirm_password', "Passwords don't match.")
        ]
    )

    confirm_password = fields.PasswordField('Confirm password')


# Chains
set_new_password_chains = ChainMgr()

# GET
set_new_password_chains['get'] = Chain([
    'config',
    'init_form',
    'apply_password_rules',
    'get_user',
    'decorate',
    'render_template'
])

# POST
set_new_password_chains['post'] = Chain([
    'config',
    'init_form',
    'apply_password_rules',
    'get_user',
    'validate',
    [
        [
            'set_password',
            'sign_in',
            'redirect'
        ],
        [
            'decorate',
            'render_template'
        ]
    ]
])

# Define the links
set_new_password_chains.set_link(
    factories.config(form_cls=SetNewPasswordForm)
)
set_new_password_chains.set_link(factories.init_form())
set_new_password_chains.set_link(apply_password_rules())
set_new_password_chains.set_link(factories.validate())
set_new_password_chains.set_link(
    factories.render_template('set_new_password.html')
)
set_new_password_chains.set_link(factories.redirect('dashboard'))

@set_new_password_chains.link
def decorate(state):
    state.decor = utils.base_decor(state.manage_config, 'set_new_password')
    state.decor['title'] = 'Set new password'

@set_new_password_chains.link
def get_user(state):
    """
    Get the user who is setting a new password based on their reset token.
    """
    user_cls = state.manage_config.frame_cls
    prefix = user_cls.get_settings_prefix()

    reset_password_url = flask.url_for(
        state.manage_config.get_endpoint('reset_password')
    )

    reset_token = flask.request.values.get('reset_token', '').strip()
    if not reset_token:

        # The token was provided, notify the user and redirect them to request
        # a new reset link.
        flask.flash(
            'Not a valid reset password link, no reset token provided.',
            'warning'
        )
        return flask.redirect(reset_password_url)

    # Get the user by their reset token
    user = user_cls.one(Q.password_reset_token == reset_token)
    state[state.manage_config.var_name] = user

    if not user:

        # No user found for the reset token, notify the user and redirect
        # them to request a new reset link.
        flask.flash(
            (
                'Not a valid reset password link, the reset token does not '
                'exist.'
            ),
            'warning'
        )
        return flask.redirect(reset_password_url)

    reset_lifespan = flask\
        .current_app\
        .config[f'{prefix}_PASSWORD_RESET_LIFESPAN']
    ellapsed = datetime.utcnow() - user.password_reset_requested

    if ellapsed > reset_lifespan:

        # The invite has expired, notify the user and redirect them to request
        # a new invite.
        flask.flash(
            (
                'It appears this reset password link has expired please  '
                'request a new one.'
            ),
            'warning'
        )
        return flask.redirect(reset_password_url)

    if user_cls.is_locked_out(user.email):

        # The user is currently locked out for too many failed attempts to
        # sign-in attempts, notify the user and redirect them to the set new
        # password page.
        lockout = flask.current_app.config[f'{prefix}_FAILED_SIGN_IN_LOCKOUT']
        minutes = round(lockout.total_seconds() / 60)
        flask.flash(
            f'Your account is locked out due to too many failed sign '
            f'in attempts, please try again in {minutes} minutes.',
            'error'
        )
        return flask.redirect(
            flask.url_for(state.manage_config.get_endpoint('sign_in'))
        )

@set_new_password_chains.link
def set_password(state):
    """Set a new password for a user"""
    user_cls = state.manage_config.frame_cls
    user = state[state.manage_config.var_name]

    # Set the user's password
    user.password = state.form.data['password']
    user.update()

    # Log the change
    entry = ChangeLogEntry({
        'type': 'NOTE',
        'documents': [user],
        'user': user
    })
    entry.add_note('Password reset')
    entry.insert()

    # Unset the invite and reset password token fields so the reset link and
    # any outstanding invite link can't be (re)used.
    user_cls.get_collection().update(
        {'_id': user._id},
        {'$unset': {'invite_token': '', 'password_reset_token': ''}}
    )

@set_new_password_chains.link
def sign_in(state):
    """Sign the user in after they have reset their password"""
    user_cls = state.manage_config.frame_cls
    user = state[state.manage_config.var_name]

    user.sign_in(force_new_token=True)
    flask.session[user_cls.get_session_token_key()] = user.session_token



"""
Update a user.
"""

import flask
from manhattan.forms import fields, validators
from manhattan.manage.views import factories, generic
from manhattan.users.utils import apply_password_rules, password_validator
from manhattan.users.views.add import AddForm
from mongoframes import Q

__all__ = ['update_chains']


# Forms

class UpdateForm(AddForm):

    new_password = fields.PasswordField(
        'Password',
        validators=[
            password_validator,
            validators.EqualTo('confirm_password', "Passwords don't match")
        ]
    )

    confirm_password = fields.PasswordField('Confirm password')

    def validate_email(form, field):
        """Validate email is unique"""
        user_cls = form._user_cls

        # The email address can stay the same for the user (irrespective of
        # case).
        if field.data.lower() == form.obj.email_lower:
            return

        user = user_cls.one(
            Q.email_lower == field.data.lower(),
            projection={'_id': True}
        )
        if user:
            raise validators.ValidationError(
                'This email address is already registered.'
            )


# Chains
update_chains = generic.update.copy()
update_chains.insert_link('init_form', 'apply_password_rules', after=True)

# Factory overrides
update_chains.set_link(
    factories.config(
        form_cls=UpdateForm,
        projection=None
    )
)
update_chains.set_link(apply_password_rules())

# Custom overrides

@update_chains.link
def init_form(state):
    # Initialize the form
    state.form = state.form_cls(
        flask.request.form,
        obj=state[state.manage_config.var_name]
    )

    # Store the user class against the form
    state.form._user_cls = state.manage_config.frame_cls

@update_chains.link
def build_form_data(state):
    # Switch the `new_password` field with password
    state.form_data = state.form.data
    state.form_data['password'] = state.form_data.pop('new_password')

@update_chains.link
def update_document(state):
    generic.update['get'].super(state)

    # If a new password has been set for the user sign them out
    if state.form_data.get('password'):
        state.manage_config.frame_cls.get_collection().update(
            {'_id': state[state.manage_config.var_name]._id},
            {'$unset': {'session_token': ''}}
        )



"""
Allow a user to change their password, view, enable and disable their
multi-factor authentication and
"""

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.forms import BaseForm, fields, validators
from manhattan.manage.views import factories
from manhattan.nav import Nav, NavItem
from manhattan.users.utils import apply_password_rules, password_validator
from mongoframes import Q, SortBy

__all__ = ['security_chains']


# Forms

class ChangeMyPasswordForm(BaseForm):

    current_password = fields.PasswordField(
        'Current password',
        [validators.Required()]
    )

    new_password = fields.PasswordField(
        'New password',
        [
            validators.Required(),
            password_validator,
            validators.EqualTo('confirm_password', "Passwords don't match.")
        ]
    )

    confirm_password = fields.PasswordField('Confirm password')

    def validate_current_password(form, field):
        """
        The user must specify their existing password to change their existing
        one.
        """
        user_cls = form._user_cls
        user = getattr(flask.g, user_cls.get_g_key())

        if not field.data:
            return

        if not user.password_eq(field.data):
            raise validators.ValidationError('Password is incorrect.')


# Chains
security_chains = ChainMgr()

# GET
security_chains['get'] = Chain([
    'config',
    'authenticate',
    'init_form',
    'apply_password_rules',
    'get_sessions',
    'decorate',
    'render_template'
])

# POST
security_chains['post'] = Chain([
    'config',
    'authenticate',
    'init_form',
    'apply_password_rules',
    'validate',
    [
        [
            'change_password',
            'redirect'
        ], [
            'get_sessions',
            'decorate',
            'render_template'
        ]
    ]
])

# Define the links
security_chains.set_link(
    factories.config(
        form_cls=ChangeMyPasswordForm,
        open_user_nav=True
    )
)
security_chains.set_link(factories.authenticate())
security_chains.set_link(apply_password_rules())
security_chains.set_link(factories.validate())
security_chains.set_link(factories.render_template('security.html'))
security_chains.set_link(factories.redirect('security'))

@security_chains.link
def decorate(state):
    factories.decorate('security')(state)

    # Modify the breadcrumb
    state.decor['breadcrumbs'] = Nav.local_menu()
    state.decor['breadcrumbs'].add(NavItem('Security', ''))

@security_chains.link
def init_form(state):
    # Initialize the form
    state.form = state.form_cls(flask.request.form)

    # Store the user class against the form
    state.form._user_cls = state.manage_config.frame_cls

@security_chains.link
def change_password(state):
    """Change the current user's password"""
    user_cls = state.manage_config.frame_cls
    user = getattr(flask.g, user_cls.get_g_key())

    # Switch the `new_password` field with password
    form_data = state.form.data
    form_data['password'] = form_data.pop('new_password')

    # Change the password
    user.logged_update(user, form_data)

    # Force a change of session token
    user.sign_in(force_new_token=True)
    flask.session[user_cls.get_session_token_key()] = user.session_token

    # Notify the user that their new password has been set
    flask.flash('New password set.')

@security_chains.link
def get_sessions(state):
    """Get a list of the current user's sessions"""
    user_cls = state.manage_config.frame_cls
    user = getattr(flask.g, user_cls.get_g_key())

    state.user_sessions = user_cls.get_session_cls().many(
        Q.user == user,
        sort=SortBy(Q.last_accessed.desc)
    )



"""
Disable multi-factor authentication (MFA) for the user.
"""

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.forms import BaseForm, fields, validators
from manhattan.manage.views import factories, utils
from manhattan.nav import NavItem

__all__ = ['disable_chains']


# Forms

class DisableForm(BaseForm):

    current_password = fields.PasswordField(
        'Password',
        [validators.Required()]
    )

    def validate_current_password(form, field):
        """
        The user must specify their existing password to disable multi-factor
        authentication.
        """

        if not field.data:
            return

        if not form.obj.password_eq(field.data):
            raise validators.ValidationError('Password is incorrect.')


# Chains
disable_chains = ChainMgr()

# GET
disable_chains['get'] = Chain([
    'config',
    'authenticate',
    'init_form',
    'decorate',
    'render_template'
])

# POST
disable_chains['post'] = Chain([
    'config',
    'authenticate',
    'init_form',
    'validate',
    [
        [
            'disable_mfa',
            'redirect'
        ], [
            'decorate',
            'render_template'
        ]
    ]
])

# Define the links
disable_chains.set_link(
    factories.config(
        form_cls=DisableForm,
        open_user_nav=True
    )
)
disable_chains.set_link(factories.authenticate())
disable_chains.set_link(factories.validate())
disable_chains.set_link(factories.render_template('mfa/disable.html'))
disable_chains.set_link(factories.redirect('security'))

@disable_chains.link
def decorate(state):
    state.decor = utils.base_decor(state.manage_config, 'mfa_disable')
    state.decor['title'] = 'Disable 2-factor authentication'

    state.decor['breadcrumbs'].add(
        NavItem('Security', state.manage_config.get_endpoint('security'))
    )
    state.decor['breadcrumbs'].add(
        NavItem('Disable 2-factor authentication', '')
    )

@disable_chains.link
def init_form(state):
    user_cls = state.manage_config.frame_cls

    # Initialize the form
    state.form = state.form_cls(
        flask.request.form,
        obj=getattr(flask.g, user_cls.get_g_key())
    )

    # Store the user class against the form
    state.form._user_cls = user_cls

@disable_chains.link
def disable_mfa(state):
    """Disable multi-factor authentication for the user"""
    user_cls = state.manage_config.frame_cls
    user = getattr(flask.g, user_cls.get_g_key())

    # Disable MFA for the user
    user.logged_update(
        user,
        {
            'mfa_enabled': False,
            'mfa_otp_secret': None,
            'mfa_recovery_codes': None
        },
        'mfa_enabled',
        'mfa_otp_secret',
        'mfa_recovery_codes',
        'modified'
    )

    current_session = user.current_session
    current_session.mfa_authorized = None
    current_session.update('mfa_authorized')

    # Notify the user the MFA has been disabled
    flask.flash('2-factor authentication disabled')



import hashlib
from datetime import datetime

import flask
from manhattan.secure import random_token
from mongoframes import And, ASC, Frame, IndexModel, Q, SortBy

__all__ = ['BaseUserSession']


class BaseUserSession(Frame):
    """
    The `BaseUserSession` collection implements device tracking for users
    in order to inform them when a new device is used and reduce the risk of
    session hijacking.

    By default the data we use to create a fingerprint is pretty broad,
    basically country (if available or IP address if not) and operating
    system. This data is return by the `discover_location` and
    `discover_platform` methods respectively. To tailor these to your needs
    simply overwrite these methods.

    NOTE: This is a **base** class and should not be used directly, instead
    define a new `UserSession` class for your project that inherits from this
    class.
    """

    _fields = {

        # The user the session relates to
        'user',

        # The location the session was created from
        'location',

        # The platform the session was created with
        'platform',

        # The date/time the user last signed in with this session
        'signed_in',

        # The date/time the user last used multi-factor authorization to
        # authorize the session.
        'mfa_authorized',

        # The date/time the user last access the app using this session
        'last_accessed',

        # A fingerprint created using the location and platform used to
        # uniquely identify the device used to create the session.
        'fingerprint_hash'
    }

    _recommended_indexes = [
        IndexModel([('user', ASC), ('fingerprint_hash', ASC)], unique=True)
    ]

    # Class methods

    @classmethod
    def discover_location(cls):
        """
        Return the location for the device (e.g country, town, lat/lng, IP
        address, etc.)
        """

        # Country
        if flask.request.headers.get('X-Country'):
            return flask.request.headers.get('X-Country')

        # IP address
        raw_ip_addresses = flask.request.headers.get('X-Forwarded-For', '')
        ip_addresses = [
            ipa.strip()
            for ipa in raw_ip_addresses.split(',')
            if ipa.strip() and not ipa.strip().startswith('10.')
        ]

        if ip_addresses:
            return ip_addresses[0]

        return 'Unknown'

    @classmethod
    def discover_platform(cls):
        """
        Return the platform for the device (e.g the operating system, browser
        user agent, manufacturer, etc.)
        """
        return flask.request.user_agent.platform or 'Unknown'

    @classmethod
    def get_current_session(cls, user_id, projection=None):
        """Return the current session"""
        return cls.one(
            And(
                Q.user == user_id,
                Q.fingerprint_hash == cls.make_fingerprint()['hash']
            ),
            projection=projection
        )

    @classmethod
    def sign_in(cls, user_id):
        """
        Sign-in and return a user session and session status for the given
        user.

        We check to see if there is an an existing session for the device
        the user is signing in with or if this is a new device they
        haven't signed in with before.

        The session status returned determines if the session is a new (True)
        or existing (False) session, e.g:

            user_session, session_is_new = MyUserSessionClass.sign_in(...)
        """

        # Get a fingerprint for the session
        fingerprint = cls.make_fingerprint()

        # Check to see if a matching session already exists
        user_session = cls.one(
            And(
                Q.user == user_id,
                Q.fingerprint_hash == fingerprint['hash']
            )
        )

        new_session = False
        now = datetime.utcnow()
        if user_session:

            # Update the existing session with the new signed in date/time
            # and clear the Multi-factor authorization (MFA) authorized flag
            # to ensure MFA enabled user are required to authorize with MFA on
            # sign in.
            user_session.signed_in = now
            user_session.last_accessed = now
            user_session.mfa_authorized = None
            user_session.update(
                'signed_in',
                'last_accessed',
                'mfa_authorized'
            )

        else:

            # Create a new session (for a new device)
            user_session = cls(
                user=user_id,
                location=fingerprint['location'],
                platform=fingerprint['platform'],
                signed_in=now,
                last_accessed=now,
                mfa_authorized=None,
                fingerprint_hash=fingerprint['hash']
            )
            user_session.insert()
            new_session = True

            # Check if the maximum number of sessions (devices) for this user
            # has been exceeded.
            user_session_count = cls.count(Q.user == user_id)
            max_user_sessions = flask.current_app.config.get(
                'USER_MAX_SESSIONS',
                12
            )

            if user_session_count > max_user_sessions:

                # If there are more sessions (devices) than the maximum
                # allowed select the oldest session(s) based on their last
                # accessed date/time and remove them.
                remove_count = max_user_sessions - user_session_count
                sessions_to_remove = cls.many(
                    Q.user == user_id,
                    sort=SortBy(Q.last_accessed),
                    limit=remove_count
                )

                # Remove old sessions
                for session_to_remove in sessions_to_remove:
                    session_to_remove.delete()

        return (user_session, new_session)

    @classmethod
    def verify_access(cls, user_id, session_lifespan):
        """
        Verify a user is using a known session (device) and update the last
        accessed date/time for the user.

        The method returns True if a session was successfully found and
        updated, and False if the session (device) was not recognized for this
        user.
        """

        # Get a fingerprint for the session
        fingerprint = cls.make_fingerprint()

        # Attempt to update the last accessed value for the session
        cutoff = datetime.utcnow() - session_lifespan
        r = cls.get_collection().update_one(
            And(
                Q.user == user_id,
                Q.fingerprint_hash == fingerprint['hash'],
                Q.last_accessed >= cutoff
            ).to_dict(),
            {'$set': {'last_accessed': datetime.utcnow()}}
        )

        # Return True if we successfully updated a session, otherwise return
        # False.
        return r.matched_count > 0

    @classmethod
    def make_fingerprint(cls):
        """
        Generate a fingerprint for a session based on information about the
        user's device.
        """

        location = cls.discover_location()
        platform = cls.discover_platform()

        # Build the fingerprint
        h = hashlib.sha256()
        h.update(location.encode('utf-8'))
        h.update(platform.encode('utf-8'))

        return {
            'location': location,
            'platform': platform,
            'hash': h.hexdigest()
        }



"""
Allow users to sign-in.
"""

import os

import flask
from manhattan.chains import Chain, ChainMgr
from manhattan.forms import BaseForm, fields, validators
from manhattan.manage.views import factories, utils
from manhattan.secure import safe_redirect_url
from mongoframes import Q

__all__ = ['sign_in_chains']


# Forms

class SignInForm(BaseForm):

    next = fields.HiddenField()

    email = fields.StringField(
        'Email',
        [validators.Required()]
    )

    password = fields.PasswordField(
        'Password',
        [validators.Required()]
    )

    remember_me = fields.BooleanField('Remember me')

    def validate_email(form, field):
        """Validate the login details"""

        user_cls = form._user_cls

        if not field.data:
            return

        if not form.password.data:
            return

        # Check that the email has not been locked out for too many failed
        # attempts.
        if user_cls.is_locked_out(field.data):
            raise validators.ValidationError('To many failed attempts.')

        # Validate we can find the user
        user = user_cls.one(
            Q.email_lower == form.email.data.lower(),
            projection={
                'password_hash': True,
                'password_salt': True
            }
        )
        if not user:
            raise validators.ValidationError('Invalid email or password.')

        # Check the user has set a password
        if not user.password_hash:
            raise validators.ValidationError('Invalid email or password.')

        # Check the password matches
        if not user.password_eq(form.password.data):
            raise validators.ValidationError('Invalid email or password.')


# Chains

sign_in_chains = ChainMgr()

# GET
sign_in_chains['get'] = Chain([
    'config',
    'authenticate',
    'init_form',
    'decorate',
    'render_template'
])

# POST
sign_in_chains['post'] = Chain([
    'config',
    'authenticate',
    'init_form',
    'validate',
    [
        [
            'sign_in',
            'notify',
            'redirect'
        ],
        [
            'log_failed_attempts',
            'decorate',
            'render_template'
        ]
    ]
])

# Define the links
sign_in_chains.set_link(factories.config(form_cls=SignInForm))
sign_in_chains.set_link(factories.validate())
sign_in_chains.set_link(factories.render_template('sign_in.html'))

@sign_in_chains.link
def authenticate(state):
    user_cls = state.manage_config.frame_cls

    if flask.g.get(user_cls.get_g_key()) \
            and user_cls.get_session_token_key() in flask.session:

        # User is already logged in redirect them to the dashboard
        return flask.redirect(
            flask.url_for(state.manage_config.get_endpoint('dashboard'))
        )

@sign_in_chains.link
def decorate(state):
    state.decor = utils.base_decor(state.manage_config, 'sign_in')
    state.decor['title'] = 'Sign-in'

@sign_in_chains.link
def init_form(state):
    # Initialize the form
    state.form = state.form_cls(
        flask.request.form,
        next=flask.request.args.get('next', '')
    )

    # Store the user class against the form
    state.form._user_cls = state.manage_config.frame_cls

@sign_in_chains.link
def log_failed_attempts(state):
    """Log failed attempts"""
    user_cls = state.manage_config.frame_cls
    prefix = user_cls.get_settings_prefix()

    if not (state.form.email.data and state.form.password.data):
        # We only log an attempt as failed if both an email and password were
        # entered.
        return

    # Log the failed attempt
    cache = flask.current_app.config[f'{prefix}_FAILED_SIGN_IN_CACHE']
    attempt_key = user_cls.get_sign_in_attempt_key(state.form.email.data)
    attempts = cache.get(attempt_key) or 0

    # Record the failed attempt
    cache.set(
        attempt_key,
        attempts + 1,
        flask.current_app.config[f'{prefix}_FAILED_SIGN_IN_LOCKOUT']
    )

    if user_cls.is_locked_out(state.form.email.data):

        # Notify the user that their account is locked out
        flask.session['_flashes'] = []

        lockout = flask.current_app.config[f'{prefix}_FAILED_SIGN_IN_LOCKOUT']
        minutes = round(lockout.total_seconds() / 60)

        flask.flash(
            'Your account is locked due to too many failed sign in '
            f'attempts, please try again in {minutes} minutes.',
            'error'
        )

        return flask.redirect(
            flask.url_for(state.manage_config.get_endpoint('sign_in'))
        )

@sign_in_chains.link
def sign_in(state):
    """
    Sign a user into the application.

    This link adds `user_session` (an instance of the users current session)
    and `new_session` (a flag indicating if the current session represents a
    new session (device)) to the state.
    """
    user_cls = state.manage_config.frame_cls

    # Sign the user in
    user = user_cls.one(
        Q.email_lower == state.form.data['email'].lower()
    )
    state[state.manage_config.var_name] = user
    user_session, new_session = user.sign_in()
    state.session = user_session
    state.new_session = new_session

    # Set the session token as a secure cookie
    flask.session[user_cls.get_session_token_key()] = user.session_token

    # If the user ticked the `remember_me` option then set the secure cookie
    # to be permenant (e.g to remain once the user closes their browser
    # client).
    if state.form.data['remember_me']:
        flask.session.permanent = True

@sign_in_chains.link
def notify(state):
    """
    Nofity the user if a new session (device) was registered when signing them
    in.
    """

    if state.new_session:

        env = flask.current_app.jinja_env
        template_path = os.path.join(
            state.manage_config.template_path,
            'emails/new_device.html'
        )
        user = state[state.manage_config.var_name]

        flask.current_app.manage.send_email(
            [user.email],
            'A new device was used to sign-in',
            template_path,
            global_vars={
                'user': user.to_json_type(),
                'user_session': state.session.to_json_type()
            },
            template_map={
                'new_device': env.loader.get_source(
                    env,
                    'manhattan/users/emails/new_device.html'
                )[0],
                template_path: env.loader.get_source(
                    env,
                    template_path
                )[0]
            }
        )

@sign_in_chains.link
def redirect(state):
    # Redirect the user to either the URL they were trying to get to or the
    # dashboard.

    next_url = safe_redirect_url(state.form.data['next'] or '')

    if not next_url:
        # If no valid next URL is provided redirect the user to the dashboard
        next_url = flask.url_for(
            state.manage_config.get_endpoint('dashboard')
        )

    return flask.redirect(next_url)
